#############################################################################
##
#W FundamentalDomainBieberbach.gi 			 HAPcryst package		 Marc Roeder
##
##  

##
#H @(#)$Id: FundamentalDomainBieberbach.gi, v 0.1.4 2007/10/17 14:22:33 gap Exp $
##
#Y	 Copyright (C) 2006 Marc Roeder 
#Y 
#Y This program is free software; you can redistribute it and/or 
#Y modify it under the terms of the GNU General Public License 
#Y as published by the Free Software Foundation; either version 2 
#Y of the License, or (at your option) any later version. 
#Y 
#Y This program is distributed in the hope that it will be useful, 
#Y but WITHOUT ANY WARRANTY; without even the implied warranty of 
#Y MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#Y GNU General Public License for more details. 
#Y 
#Y You should have received a copy of the GNU General Public License 
#Y along with this program; if not, write to the Free Software 
#Y Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
##
Revision.("/home/roeder/gap/HAPcryst/HAPcryst/lib/FundamentalDomainBieberbach_gi"):=
	"@(#)$Id: FundamentalDomainBieberbach.gi, v 0.1.4 2007/10/17   14:22:33  gap Exp $";
#############################################################################
##
#O FundamentalDomainBieberbachGroupNC(<group>)
##
InstallMethod(FundamentalDomainBieberbachGroupNC,
        [IsGroup],
        function(group)
    local center;
    center:=List([1..Size(Representative(group))-1],i->0);
    return FundamentalDomainBieberbachGroupNC(center,group);
end);

#############################################################################
##
#O FundamentalDomainBieberbachGroup(<group>)
##
InstallMethod(FundamentalDomainBieberbachGroup,
        [IsGroup],
        function(group)
    local center;
    center:=List([1..Size(Representative(group))-1],i->0);
    return FundamentalDomainBieberbachGroup(center,group);
end);


#############################################################################
##
#O FundamentalDomainBieberbachGroupNC(<center>, <group>)
##
InstallMethod(FundamentalDomainBieberbachGroupNC,
        [IsVector,IsGroup],
        function(center,group)
    local gram;

    gram:=GramianOfAverageScalarProductFromFiniteMatrixGroup(PointGroup(group));
    return FundamentalDomainBieberbachGroupNC(center,group,gram);
end);


#############################################################################
##
#O FundamentalDomainBieberbachGroup(<center>, <group>)
##
InstallMethod(FundamentalDomainBieberbachGroup,
        [IsVector,IsGroup],
        function(center,group)
    local gram;

    gram:=GramianOfAverageScalarProductFromFiniteMatrixGroup(PointGroup(group));
    return FundamentalDomainBieberbachGroup(center,group,gram);
end);


#############################################################################
##
#O FundamentalDomainBieberbachGroup(<center>, <group>, <gram>)
##
InstallMethod(FundamentalDomainBieberbachGroup,
        [IsVector,IsGroup,IsMatrix],
        function(center,group,gram)
    local   dim;
    if not (IsStandardSpaceGroup(group) and IsAffineCrystGroupOnRight(group))
       then
        Error("group must be a StandardSpaceGroup acting on right");
    elif not IsAlmostBieberbachGroup(Image(IsomorphismPcpGroup(group)))
      then
        Error("group must be a Bieberbach group");
    fi;
    dim:=Size(center);
    if not (dim=Size(gram) and DimensionOfMatrixGroup(group)=dim+1)
       then
        Error("dimensions don't match");
    fi;
    return FundamentalDomainBieberbachGroupNC(center,group,gram);
end);
    
    

#############################################################################
##
#O FundamentalDomainBieberbachGroupNC(<center>, <group>, <gram>)
##
InstallMethod(FundamentalDomainBieberbachGroupNC,
        [IsVector,IsGroup,IsMatrix],
        function(center,group,gram)
    local   initialInequalities,  satisfiesAllInequalities,  
            newInequalitiesEuclidian,  newInequalities,  dim,  
            isEuclidian,  orbitpart,  partialFD,  newinequalities,  
            oldinequalities,  donevertices,  rsquare,  allvertices,  
            box,  vertices,  vertexiterator,  inequalities,  v,  new;
    
    ###################
    ##  A lot of helper functions:
    ###################


    initialInequalities:=function(center,orbitpart,gram)
        local   dim,  directions,  points,  ineqs;

        dim:=Size(center);
        directions:=Union(IdentityMat(dim),-IdentityMat(dim));
        points:=[];
        points:=Set(ShiftedOrbitPart(center,orbitpart));
        UniteSet(points,center+directions);
        #    UniteSet(points,Union(List(directions,d->d+points)));
        if gram<>IdentityMat(dim)
           then
            ineqs:=Set(points,
                       i->BisectorInequalityFromPointPair(center,i,gram));
        else
            ineqs:=Set(points,
                       i->BisectorInequalityFromPointPair(center,i));
        fi;
        RemoveSet(ineqs,fail);
        return ineqs;
    end;


 
   satisfiesAllInequalities:=function(point,ineqs)
        return ForAll(ineqs,i->WhichSideOfHyperplane(point,i) in [0,1]);
    end;



    newInequalitiesEuclidian:=function(vertex,center,ineqs,vertices,box,group)
        local   dim,   affinevertex,  returnlist,  g,  
                affimage,  image,  trans,  t,  gl,  gli,  gt;

        dim:=Size(vertex);
        affinevertex:=Concatenation(vertex,[1]);
        returnlist:=[];
        for g in PointGroupRepresentatives(group)
          do
            affimage:=affinevertex*g;
            image:=affimage{[1..dim]};
            ###
            trans:=TranslationsToBox(image,box);
            #trans:=TranslationsToOneCubeAroundCenter(image,center);
            for t in trans 
              do
                if not image+t in vertices 
                   then
                    if satisfiesAllInequalities(image+t,ineqs)
                       then
                        gl:=LinearPartOfAffineMatOnRight(g);
                        gli:=Inverse(gl);
                        gt:=g[dim+1]{[1..dim]};
                        Add(returnlist,center*gl+gt+t);
                        Add(returnlist,(center-gt-t)*gli);
                    fi;
                fi;
            od;
        od;
        returnlist:=Set(returnlist);
        RemoveSet(returnlist,center);
        returnlist:=Set(returnlist,i->BisectorInequalityFromPointPair(center,i));
        return Filtered(returnlist,i->ForAny(vertices,v->WhichSideOfHyperplane(v,i)=-1));
    end;


    newInequalities:=function(vertex,center,ineqs,vertices,box,group,gram)
        local   dim,   affinevertex,  returnlist,  g,  
                affimage,  image,  trans,  t,  gl,  gli,  gt;

        dim:=Size(vertex);
        affinevertex:=Concatenation(vertex,[1]);
        returnlist:=[];
        for g in PointGroupRepresentatives(group)
          do
            affimage:=affinevertex*g;
            image:=affimage{[1..dim]};
            ###
            trans:=TranslationsToBox(image,box);
            #trans:=TranslationsToOneCubeAroundCenter(image,center);
            for t in trans 
              do
                if not image+t in vertices 
                   then
                    if satisfiesAllInequalities(image+t,ineqs)
                       then
                        gl:=LinearPartOfAffineMatOnRight(g);
                        gli:=Inverse(gl);
                        gt:=g[dim+1]{[1..dim]};
                        Add(returnlist,center*gl+gt+t);
                        Add(returnlist,(center-gt-t)*gli);
                    fi;
                fi;
            od;
        od;
        returnlist:=Set(returnlist);
        RemoveSet(returnlist,center);
        returnlist:=Set(returnlist,i->BisectorInequalityFromPointPair(center,i,gram));
        return Filtered(returnlist,i->ForAny(vertices,v->WhichSideOfHyperplane(v,i)=-1));
    end;
    
    
    ###################
    ##  helper functions done. Let's go:
    ###################
    
    dim:=Size(center);
    isEuclidian:=(gram=IdentityMat(dim));
    orbitpart:=OrbitStabilizerInUnitCubeOnRight(group,center).orbit;;
# First, generate the initial polyhedron:
    partialFD:=CreatePolymakeObject("partialFD",POLYMAKE_DATA_DIR);
    newinequalities:=initialInequalities(center,orbitpart,gram);;
    AppendInequalitiesToPolymakeObject(partialFD,newinequalities);
    Polymake(partialFD,"VERTICES FACETS");
    newinequalities:=[];
    oldinequalities:=[];
    donevertices:=[];
    rsquare:=1;
    
    repeat
        allvertices:=AsSet(Polymake(partialFD,"VERTICES"));
        ##
        box:=List([1..dim],i->[Minimum(List(allvertices,v->v[i])),
                     Maximum(List(allvertices,v->v[i]))]
                 );
        ##

        vertices:=Filtered(Polymake(partialFD,"VERTICES"),v->not v in donevertices);
        if vertices=[] 
           then 
            Info(InfoHAPcryst,1,"out of vertices");
            return partialFD;
        fi;
        vertexiterator:=Iterator(vertices);
        inequalities:=Set(Polymake(partialFD,"FACETS"));
        Info(InfoHAPcryst,2,"v: ",Size(vertices)," f:", Size(inequalities),"  \c");
        newinequalities:=[];
        if isEuclidian
           then
            repeat
                v:=NextIterator(vertexiterator);
                AddSet(donevertices,v);
                if satisfiesAllInequalities(v,newinequalities)
                   then
                    new:= newInequalitiesEuclidian(v,center,inequalities,
                                  allvertices,box,group);
                    UniteSet(newinequalities,new);
                    UniteSet(inequalities,new);
                fi;
            until IsDoneIterator(vertexiterator) or Size(newinequalities)>200;
        else
            repeat
                v:=NextIterator(vertexiterator);
                AddSet(donevertices,v);
                if satisfiesAllInequalities(v,newinequalities)
                   then
                    new:= newInequalities(v,center,inequalities,
                                  allvertices,box,group,gram);
                    UniteSet(newinequalities,new);
                    UniteSet(inequalities,new);
                fi;
            until IsDoneIterator(vertexiterator) or Size(newinequalities)>200;
        fi;

        Info(InfoHAPcryst,2,"ne: ",Size(newinequalities));
        ClearPolymakeObject(partialFD);
        AppendInequalitiesToPolymakeObject(partialFD,inequalities);;
        Polymake(partialFD,"VERTICES FACETS");
    until newinequalities=[];
    Info(InfoHAPcryst,2,"out of inequalities");
    return partialFD;
end);

