#############################################################################
##
#W PartialHasseDiagramByGroup.gi 			 HAPcryst package		 Marc Roeder
##
##  

##
#H @(#)$Id: PartialHasseDiagramByGroup.gi, v 0.1.1 2007/08/23 18:08:51 gap Exp $
##
#Y	 Copyright (C) 2006 Marc Roeder 
#Y 
#Y This program is free software; you can redistribute it and/or 
#Y modify it under the terms of the GNU General Public License 
#Y as published by the Free Software Foundation; either version 2 
#Y of the License, or (at your option) any later version. 
#Y 
#Y This program is distributed in the hope that it will be useful, 
#Y but WITHOUT ANY WARRANTY; without even the implied warranty of 
#Y MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#Y GNU General Public License for more details. 
#Y 
#Y You should have received a copy of the GNU General Public License 
#Y along with this program; if not, write to the Free Software 
#Y Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
##
Revision.("/home/roeder/gap/HAPcryst/HAPcryst/lib/PartialHasseDiagramByGroup_gi"):=
	"@(#)$Id: PartialHasseDiagramByGroup.gi, v 0.1.1 2007/08/23   18:08:51  gap Exp $";

##################################################
#
#
InstallMethod(PartialHasseDiagramByGroup,
        [IsRecord,IsDenseList,IsInt],
        function(poly,GLOBALpointGroupReps,maxdim)

    local   vector2vertexindex,  vectorlist2vertexindexlist,  
            vertexindex2vector,  vertexindexlist2vectorlist,  
            liesBetweenPoints,  verticesForFaceByProjections,  
            entryForFaceFromKnownOrbits,  addPointToHasseDiagram,  
            addImagePointToHasseDiagram,  reformatedHasseDiagram,  
            unknownFaces,  orbitPartInVertices,  
            findFaceInKnownOrbits,  addOrbitToKnownOnes,  
            oneStepOfHasseDiagramByFace,  
            GLOBALmoduleGeneratorsAsFaces,  GLOBALknownOrbits,  
            basislist,  GLOBALvertexlist,  GLOBALvertexindexdata,  
            indices,  GLOBALrawHasseDiagram,  vertexset,  vertexreps,  
            vertex,  vertexindex,  orbit,  vertexposition,  
            imagevertex,  acRep,  newfaces,  projections,  proj,  
            dimension,  facestodo,  face,  next;

    ###############
    # these are the only functions that mess around with the global 
    # vertex lists.
    #
    vector2vertexindex:=function(vector)
        local   pos;
        pos:=PositionSet(GLOBALvertexindexdata.vertexlist,vector);
        if pos=fail 
           then
            return fail;
        else
            return GLOBALvertexindexdata.indices[pos];
        fi;
    end;

    vectorlist2vertexindexlist:=function(vectorlist)
        local   returnlist,  i,  index;

        returnlist:=List(vectorlist,i->0);
        for i in [1..Size(vectorlist)]
          do
            index:=vector2vertexindex(vectorlist[i]);
            if index=fail
               then
                return fail;
            else
                returnlist[i]:=index;
            fi;
        od;
        return returnlist;
    end;


    vertexindex2vector:=function(index)
        return GLOBALvertexlist[index];
    end;

    vertexindexlist2vectorlist:=function(indexlist)
        return GLOBALvertexlist{indexlist};
    end;

    ##############################

    liesBetweenPoints:=function(vector,pointpair)
        #does one step of Gauss elimination. Wow!
        local   difference,  vdifference,  pos;

        if vector in pointpair
           then 
            return true;
        fi;
        difference:=pointpair[2]-pointpair[1];
        vdifference:=vector-pointpair[1];
        pos:=1;
        while pos<Size(difference) and difference[pos]=0 and vdifference[pos]=0
          do 
            pos:=pos+1;
        od;
        if difference[pos]=0
           then 
            return false;
        elif (vdifference[pos]/difference[pos])*difference=vdifference
          then 
            return true;
        else
            return false;
        fi;
    end;


    ##############################
    # projections is a record containing 
    #  .lookupTable of faces and their respective vertices 
    #  .proj the list of projections (again records)
    #
    #  returns a list of vectors!
    #
    verticesForFaceByProjections:=function(face,facedimension,projections)
        # a face is just a point in <facedimension> dimensions!
        # this does not check if <face> is really a face (ie is a vertex in
        # the projected polygon).
        local   knowndataposition,  proj,  facepreimage,  
                faceextremepoints,  verticesforthisface,  facepoints;

        if facedimension=0 #just for convenience...
           then
            return [face];
        fi;
        if not IsBound(projections.lookupTable[facedimension]) #projections.lookupTable[facedimension]=[]
           then
            knowndataposition:=fail;
        else
            knowndataposition:=PositionProperty(projections.lookupTable[facedimension],i->i[1]=face);
        fi;
        if knowndataposition<>fail
           then
            return projections.lookupTable[facedimension][knowndataposition][2];
        fi;
        proj:=projections.proj[facedimension].projection;
        facepreimage:=First(proj,i->i[2]=face)[1];
        faceextremepoints:=[facepreimage,projections.proj[facedimension].conepoint];
        if facedimension=1
           then
            verticesforthisface:=Set(faceextremepoints);
        else
            facepoints:=Filtered(List(proj,i->i[1]),
                                j->liesBetweenPoints(j,faceextremepoints)
                                );
            Add(facepoints,projections.proj[facedimension].conepoint);
            verticesforthisface:=Union(List(facepoints,i->verticesForFaceByProjections(i,facedimension-1,projections)));
        fi;
        if not IsBound(projections.lookupTable[facedimension])
           then
            projections.lookupTable[facedimension]:=[];
        fi;
        AddSet(projections.lookupTable[facedimension],[face,verticesforthisface]);
        return verticesforthisface;
    end;
    
    
    ##############################
    # returns the position of face in the Hasse diagram if it is already known
    # returns fail if face is not contained in any known orbit.
    # Otherwise, the new entry for the Hasse diagram is returned.
    # That is, a pair [vertices,[representativeIndex,matrixTakingRepresentativeToFace]]
    #
    entryForFaceFromKnownOrbits:=function(vertexindices,facedim)
        local   repPosition,  rep,  vertexvectors,  repvectors,  
                grouprep;
        
        repPosition:=findFaceInKnownOrbits(vertexindices,facedim);
        if repPosition<>fail
           then
            rep:=GLOBALrawHasseDiagram[facedim+1][repPosition][1];
            if rep=vertexindices
               then
                return repPosition;
            fi;
            vertexvectors:=vertexindexlist2vectorlist(vertexindices);
            repvectors:=Set(vertexindexlist2vectorlist(rep));
            grouprep:=RepresentativeActionOnRightOnSets(repvectors,Set(vertexvectors),GLOBALpointGroupReps);
            return [vertexindices,[repPosition,grouprep]];
        else 
            return fail;
        fi;
    end;
    
    
    ##############################    
    # add a face to the Hasse diagram in the "three entries" form.
    # [verticesofface,invboundary,bounday]
    #
    addPointToHasseDiagram:=function(face,facedimension,projections)
        local   vertexnumber,  faceposition,  proj,  facepreimage,  
                faceextremepoints,  nextfacestoadd,  betweenfaces,  
                vectorfacevertices,  facevertices,  hasseEntry,  
                orbit,  nextface,  nextposition;

        if facedimension=0 #ie face is actually a vertex.
           then
            vertexnumber:=vector2vertexindex(face);
            faceposition:=PositionProperty(GLOBALrawHasseDiagram[1],i->i[1]=[vertexnumber]);
            if faceposition=fail #this should (must?) not happen!
               then
                Add(GLOBALrawHasseDiagram[1],[[vertexnumber],[],[]]);
                return Size(GLOBALrawHasseDiagram[1]);
            else
                return faceposition;
            fi;
        else            
            proj:=projections.proj[facedimension].projection;
            facepreimage:=First(proj,i->i[2]=face)[1];
            faceextremepoints:=[facepreimage,projections.proj[facedimension].conepoint];
            if facedimension=1
               then
                nextfacestoadd:=[[faceextremepoints,0]];
            else
                nextfacestoadd:=[[faceextremepoints,facedimension-1]];
                betweenfaces:=Filtered(List(proj,i->i[1]),
                                        j->liesBetweenPoints(j,faceextremepoints)
                                      );
                Append(nextfacestoadd,
                       List(Filtered(betweenfaces,b->b<>facepreimage),i->[i,facedimension-2]));
            fi;
#            vectorfacevertices:=Union(List(nextfacestoadd,f->
#                                        verticesForFaceByProjections(f[1],facedimension-1,projections)));
            vectorfacevertices:=Union(List(nextfacestoadd,i->i[1]));
            Apply(vectorfacevertices,v->verticesForFaceByProjections(v,facedimension-1,projections));
            vectorfacevertices:=Union(vectorfacevertices);
            facevertices:=Set(vectorlist2vertexindexlist(vectorfacevertices));
            hasseEntry:=entryForFaceFromKnownOrbits(facevertices,facedimension);
            if IsInt(hasseEntry)
               then
                faceposition:=hasseEntry;
            elif hasseEntry<>fail   #=> the face is contained in a known orbit.
               then
                Add(GLOBALrawHasseDiagram[facedimension+1],hasseEntry);
                faceposition:=Size(GLOBALrawHasseDiagram[facedimension+1]);
            else  #=> Nothing is known. The face is added in 3-entry form.
                  #   Also add the face to the module generators and the orbit.
                Add(GLOBALrawHasseDiagram[facedimension+1],[facevertices,[],[]]);
                faceposition:=Size(GLOBALrawHasseDiagram[facedimension+1]);
                orbit:=orbitPartInVertices(vectorfacevertices);
                addOrbitToKnownOnes(List(orbit,vectorlist2vertexindexlist),faceposition,facedimension);
                AddSet(GLOBALmoduleGeneratorsAsFaces[facedimension+1],facevertices);
                for nextface in nextfacestoadd
                  do
                    nextposition:=addPointToHasseDiagram(nextface[1],nextface[2],projections);
                    # add the "inverse boundary" for the point if 
                    # the point is not described by a group element (in 2-entry form)
                    if nextface[2]=facedimension-1 and not Size(GLOBALrawHasseDiagram[facedimension][nextposition])=2
                       then
                        AddSet(GLOBALrawHasseDiagram[facedimension][nextposition][2],faceposition);
                    fi;
                    # and the boundry entry for this face
                    AddSet(GLOBALrawHasseDiagram[facedimension+1][faceposition][3],nextposition);
                od;
            fi;  
            return faceposition;
        fi;
    end;


    ##############################            
    # add the face which is descibed by the vertex indices <verticesofimage>
    # to the Hasse diagram. It is represented in the form 
    # [verticesofimage,[preimageposition,representative]]
    # where preimageposition is the position of a face $f$ such that the
    # image of $f$ under <representative> is the face descibed by 
    # <verticesofimage>
    # 
    addImagePointToHasseDiagram:=function(verticesofimage,preimageposition,representative,facedimension)
        local   faceposition;

        faceposition:=PositionProperty(GLOBALrawHasseDiagram[facedimension+1],i->i[1]=verticesofimage);
        if faceposition=fail
           then
            Add(GLOBALrawHasseDiagram[facedimension+1],[verticesofimage,[preimageposition,representative]]);
        else
            GLOBALrawHasseDiagram[facedimension+1][faceposition]:=[verticesofimage,[preimageposition,representative]];
        fi;
#        preimageboundary:=GLOBALrawHasseDiagram[facedimension+1][preimageposition][3];
        
    end;

    ##############################
    # the Hasse diagram generated by addPoinToHasseDiagram
    # has the wrong form. The next function converts it to the
    # form as it is used by polymake.
    #
    # I may not use this...
    # It does not contain any group things.
    #
    reformatedHasseDiagram:=function(rawHasseDiagram)
        local   faceindices,  hasseSizes,  i,  hasse,  level;

        faceindices:=[1..Size(rawHasseDiagram)];
        hasseSizes:=List(rawHasseDiagram,Size);
        if Size(hasseSizes)=1
           then
            return rec(hasse:=rawHasseDiagram,faceindices:=[[1..hasseSizes[1]]]);
        fi;
        for i in [2..Size(hasseSizes)]
          do
            faceindices[i]:=faceindices[i-1]+[1..hasseSizes[i]];
        od;

        hasse:=List([1..Sum(hasseSizes)],i->0);

        for i in faceindices[1]
          do
            hasse[i]:=rawHasseDiagram[1][i];
            hasse[i][2]:=Set(hasse[i][2],j->j+hasseSizes[1]);
        od;
        for level in [2..Size(faceindices)-1]
          do
            for  i in faceindices[level]
              do
                hasse[i]:=rawHasseDiagram[level][i-faceindices[level][1]+1];
                hasse[i][2]:=Set(hasse[i][2],j->j+hasseSizes[level]);
                hasse[i][3]:=Set(hasse[i][3],j->j+faceindices[level-1][1]-1);
            od;
        od;
        level:=Size(faceindices);
        for i in faceindices[level]
          do
            hasse[i]:=rawHasseDiagram[level][i-faceindices[level][1]+1];
            hasse[i][3]:=Set(hasse[i][3],j->j+faceindices[level-1][1]-1);
        od;
        return rec(hasse:=hasse,faceindices:=faceindices);
    end;

    ###############

    unknownFaces:=function(facelist,facedimension)
        local   unknown,  face,  pos;

        #hasselist:=List(GLOBALrawHasseDiagram[facedimension+1]);
        #Sort(hasselist,function(x,y) x[1]<y[1];end);
        unknown:=[];
        for face in facelist
          do
            #pos:=PositionSet(List(hasselist,i->i[1]),face);
            pos:=PositionProperty(GLOBALrawHasseDiagram[facedimension+1],
                         i->i[1]=face);
            if pos=fail
               then
                Add(unknown,face);
            elif Size(GLOBALrawHasseDiagram[facedimension+1][pos])=3 
              and not face in GLOBALmoduleGeneratorsAsFaces[facedimension+1]
              then
                Add(unknown,face);
            fi;
        od;
        Sort(unknown);
        return unknown;
    end;

    ###############
    #this takes a list of vertices (vectors) and returns the part of the 
    # orbit which consists of vertex sets.
    #
    orbitPartInVertices:=function(vertexset)
        local   orbit,  allvertices,  raworbit,  set,  vertex,  trans,  
                candidate;

        if not IsSet(vertexset) and ForAll(vertexset,IsVector)
           then
            Error("vertexset must be a set of vectors");
        fi;
        orbit:=[];
        allvertices:=AsSet(GLOBALvertexlist);
        raworbit:=OrbitStabilizerInUnitCubeOnRightOnSets(vertexset,GLOBALpointGroupReps).orbit;
        for set in raworbit
          do
            for vertex in allvertices
              do
                trans:=vertex-set[1];
                candidate:=set+trans;
                if ForAll(trans,IsInt) and IsSubset(allvertices,candidate)
                   then
                    Add(orbit,candidate);
                fi;
            od;
        od;
        return Set(orbit);
    end;
    


    ###############

    findFaceInKnownOrbits:=function(vertexindices,dimension)
        local   pos;
        pos:=PositionSet(GLOBALknownOrbits.faces[dimension+1],vertexindices);
        if pos=fail
           then
            return fail;
        else
            return GLOBALknownOrbits.basislist[dimension+1][pos];
        fi;                      
    end;

    ###############

    addOrbitToKnownOnes:=function(orbitindices,basiselposition,dimension)

        Append(GLOBALknownOrbits.faces[dimension+1],orbitindices);
        Append(GLOBALknownOrbits.basislist[dimension+1],
               List([1..Size(orbitindices)],i->basiselposition));
        SortParallel(GLOBALknownOrbits.faces[dimension+1],
                GLOBALknownOrbits.basislist[dimension+1]);
    end;


    ##############################
    # <face> has to be a vector corresponding (via <projections>) to a face 
    # of dimension <facedim>.
    #
    # All faces of dimension <facedim>+1 containing <face> are calculated.
    # For each orbit under the group, a module generator is added and 
    # the Hasse diagram is extended by the corresponding element.
    # Faces of lower dimension may also be added to the Hasse diagram as
    # boundaries of the new elements.
    #
    # The method returns a list .faces of vectors representing <facedim>+1 
    # faces and a list .projections of projections containing one projection 
    # more than <projections>
    # 
    oneStepOfHasseDiagramByFace:=function(face,facedim,projections)
        local   newprojection,  extendedprojections,  facets,  
                freefacets,  addedfacets,  facet,  vectorfacet,  
                knownbasisposition,  preimagevertices,  rep,  
                facetorbit,  setindex,  facetasvector,  
                facetposition;

        #facevertices:=verticesForFaceByProjections(face,facedim,projections);
        if facedim=0
           then
            newprojection:=projections.proj[1];
            extendedprojections:=projections;
        else
            newprojection:=ProjectionOfPolygonByCone(face,projections.proj[facedim].poly);
            extendedprojections:=rec(lookupTable:=StructuralCopy(projections.lookupTable),proj:=Concatenation(projections.proj,[newprojection]));
        fi;
        if not (IsBound(newprojection.poly.VERTICES) and 
                IsBound(newprojection.poly.POINTS))
           then
            Polymake(newprojection.poly,"POINTS VERTICES");
        fi;
        facets:=List(newprojection.poly.VERTICES,v->
                     [v,verticesForFaceByProjections(v,facedim+1,
                             extendedprojections)]);
        #these facets have not yet been added to the Hasse diagram
        # or have been added "blindly":
        freefacets:=unknownFaces(List(facets,i->Set(vectorlist2vertexindexlist(i[2]))),facedim+1);
        #facets:=Set(Filtered(facets,f->f[2] in freefacets),
        #            i->vectorlist2vertexindexlist(i[2]));
        addedfacets:=[];

        while freefacets<>[]
          do
            facet:=AsSet(Remove(freefacets,1));
            vectorfacet:=AsSet(vertexindexlist2vectorlist(facet));
            knownbasisposition:=findFaceInKnownOrbits(facet,facedim+1);
            if knownbasisposition<>fail
               then
                preimagevertices:=GLOBALrawHasseDiagram[facedim+2][knownbasisposition];
                rep:=RepresentativeActionOnRightOnSets(
                             Set(vertexindexlist2vectorlist(preimagevertices[1])),
                             vectorfacet,
                             GLOBALpointGroupReps);
                addImagePointToHasseDiagram(facet,knownbasisposition,rep,facedim+1);
            else
                # now <facet> will be an new module generator.
                #

                facetorbit:=orbitPartInVertices(vectorfacet);
                for setindex in [1..Size(facetorbit)]
                  do
                    facetorbit[setindex]:=vectorlist2vertexindexlist(facetorbit[setindex]);
                    if facetorbit[setindex]=fail
                       then
                        Unbind(facetorbit[setindex]);
                    else
                        Sort(facetorbit[setindex]);
                        MakeImmutable(facetorbit[setindex]);
                    fi;
                od;
                facetasvector:=First(facets,i->i[2]=vectorfacet)[1];
                Add(addedfacets,facetasvector);
                facetposition:=addPointToHasseDiagram(
                                       facetasvector,
                                       facedim+1,
                                       extendedprojections);
#                AddSet(GLOBALmoduleGeneratorsAsFaces[facedim+2],facet);
#                addOrbitToKnownOnes(facetorbit,facetposition,facedim+1);
                RemoveSet(freefacets,facetorbit);
            fi;
        od;
        return rec(faces:=Set(addedfacets),projections:=extendedprojections);
    end;

    ###################################
    ## Now we finally generate the partial Hasse diagram!
    ###################################
    #GLOBAL VARIABLES for this function (ugly but true!, prefixed by GLOBAL)
    #
    # pointGroupReps: Representatives for the point group (passed 
    #             as an option)
    # vertexlist: List of vectors correspoding to the vertices.
    #             As returned by Polymake(poly,"VERTICES");
    #             
    # moduleGeneratorsAsFaces: A list with i^th entry a list of 
    #             i+1 faces (as list of vertex-indices). Finally, this
    #             will be the generators of the modules in the
    #             resolution.
    #
    # rawHasseDiagram: a Hasse diagram containing two sorts of entries. One
    #                  of the form [vertexindices,invboundary,boundary]
    #                  and one of the form [vertexindices,[faceindex,groupel]]
    #                  where the image of the face belonging to <faceindex.
    #                  under <groupel> is the face descibed by <vertexindices>
    #
    # knownOrbits: A record containing a lists
    #              .faces which i^th entry contains
    #              all faces which belong to some orbit of a module generator
    #              "of dimension i-1".
    #              Eeach of this lists is supposed to be sorted (ie a SET)
    #              
    #              .basislist which corresponds to the entries in .faces, but 
    #              holds the index of the basis element (in the 
    #              rawHasseDiagram[i+1])the face belongs to.
    #              The entries of .faces and .basislist have to be parallel- 
    #              sorted!
    #####################################

    GLOBALmoduleGeneratorsAsFaces:=List([1..maxdim+1],i->[]);
    GLOBALknownOrbits:=rec(faces:=List(GLOBALmoduleGeneratorsAsFaces,g->[]),
                           basislist:=List(GLOBALmoduleGeneratorsAsFaces,g->[]));
    GLOBALvertexlist:=AsList(Polymake(poly,"VERTICES"));
    GLOBALvertexindexdata:=rec(vertexlist:=List(poly.VERTICES),
                               indices:=[1..Size(GLOBALvertexlist)]);
    SortParallel(GLOBALvertexindexdata.vertexlist,
            GLOBALvertexindexdata.indices);
    MakeImmutable(GLOBALvertexindexdata);
    GLOBALrawHasseDiagram:=List([1..maxdim+1],i->[]);
    # the first entry of the Hasse diagram will contain all 
    # vertices to make the representation more consistent.
    # In particular, the positions of the vertices are identical 
    # with their name.
   GLOBALrawHasseDiagram[1]:=List([1..Size(GLOBALvertexlist)],i->[[i],[],[]]);

    vertexset:=Set(GLOBALvertexlist);
    vertexreps:=[];
    while vertexset<>[]
      do
        vertex:=Remove(vertexset,1);
        vertexindex:=vector2vertexindex(vertex);
        Add(vertexreps,vertex);
        orbit:=Union(orbitPartInVertices([vertex]));
        addOrbitToKnownOnes(Union(vectorlist2vertexindexlist(orbit),[vertexindex]),vertexindex,0);
        SubtractSet(vertexset,orbit);
        vertexposition:=addPointToHasseDiagram(vertex,0,[]);

        RemoveSet(orbit,vertex);
        AddSet(GLOBALmoduleGeneratorsAsFaces[1],[vertexindex]);
        for imagevertex in orbit
          do
            acRep:=RepresentativeActionOnRightOnSets([vertex],[imagevertex],GLOBALpointGroupReps);
            addImagePointToHasseDiagram([vector2vertexindex(imagevertex)],vertexposition,acRep,0);
        od;
    od;

    newfaces:=[];
    for vertex in vertexreps
      do
        Add(newfaces,
            rec(face:=vertex,
                      projections:=rec(lookupTable:=[],
                              proj:=[ProjectionOfPolygonByCone(vertex,poly)])
                      )
                      );
    od;

    for dimension in [0..maxdim-1]
      do
        facestodo:=newfaces;
        newfaces:=[];
        for face in facestodo
          do
            next:=oneStepOfHasseDiagramByFace(face.face,dimension,face.projections);
            Append(newfaces,List(next.faces,f->rec(face:=f,projections:=next.projections)));
        od;
    od;
    #
    #CONVERT HASSE DIAGRAM!
    #
    Error("take a look!");
    return rec(hasse:=GLOBALrawHasseDiagram,gens:=GLOBALmoduleGeneratorsAsFaces);
end);