#############################################################################
##
#W facelattice.gi 			 HAPcryst package		 Marc Roeder
##
##  

##
#H @(#)$Id: facelattice.gi, v 0.1.1 2007/08/23 18:08:51 gap Exp $
##
#Y	 Copyright (C) 2006 Marc Roeder 
#Y 
#Y This program is free software; you can redistribute it and/or 
#Y modify it under the terms of the GNU General Public License 
#Y as published by the Free Software Foundation; either version 2 
#Y of the License, or (at your option) any later version. 
#Y 
#Y This program is distributed in the hope that it will be useful, 
#Y but WITHOUT ANY WARRANTY; without even the implied warranty of 
#Y MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#Y GNU General Public License for more details. 
#Y 
#Y You should have received a copy of the GNU General Public License 
#Y along with this program; if not, write to the Free Software 
#Y Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
##
Revision.("/home/roeder/gap/HAPcryst/HAPcryst/lib/facelattice_gi"):=
	"@(#)$Id: facelattice.gi, v 0.1.1 2007/08/23   18:08:51  gap Exp $";
InstallMethod(FaceLatticeAndBoundary,
        [IsRecord,IsDenseList],
        function(IsPolymakeObject,Preps)  
    local   removeSomeFaces,  orbitDecompositionAsIndices,  
            changeHasseEntries,  reformatHD,  calculateBoundary,  dim,  
            vertices,  hasse,  codim,  timetmp,  faceOrbits,  k,  j,  
            elts;
    
    
    removeSomeFaces:=function(upfaces,faces)
        local   faceindex,  face;
        for faceindex in [1..Size(faces)]
          do
            face:=faces[faceindex];
            if not ForAny(upfaces,f->IsSubset(f,face))
               then
                Unbind(faces[faceindex]);
            fi;
        od;
    end;
    
    
    orbitDecompositionAsIndices:=function(facelist,vertexlist,pgrpreps)
        local   list,  localvertices,  returnlist,  face,  orbit;
        
        if not ForAll(facelist,IsSet)
           then
            Error("face list must be a list of sets");
        fi;
        list:=ShallowCopy(facelist);
        localvertices:=Set(vertexlist{Set(Flat(list))});
        returnlist:=[];
        while list<>[]
          do
            face:=list[1];
            orbit:=OrbitPartInVertexSetsAffineCrystGroupOnRight(Set(vertexlist{face}),localvertices,pgrpreps);
            Apply(orbit,o->Set(o,i->Position(vertexlist,i)));
            orbit:=Filtered(orbit,i->i in facelist);
            SubtractSet(list,orbit);
            Apply(orbit,i->Position(facelist,i));
            Add(returnlist,orbit);
        od;
        return returnlist;    
    end;
    
    
    
    changeHasseEntries:=function(orbits,codim,vertices,hasse,pgrpreps)
        local   genIndices,  faces,  newhasseentry,  idMat,  orbit,  
                gen,  genIndex,  upindices,  upface,  o,  entry;

        genIndices:=[];
        faces:=hasse[codim+1];
        newhasseentry:=[];
        idMat:=IdentityMat(Size(hasse));
        for orbit in orbits 
          do
            gen:=faces[orbit[1]];
            Add(newhasseentry,[gen,[]]);
            genIndex:=Size(newhasseentry);
            upindices:=Filtered([1..Size(hasse[codim])],
                               i->IsSubset(hasse[codim][i][1],gen));
            #Generate "inverse boundary" entry for new generator:
            #hasse[codim+1][genIndex]:=[hasse[codim+1][genIndex],
            #                           Set(upindices),
            #                           []];
            #update boundary for faces containing new generator:
            for upface in upindices
              do
                AddSet(hasse[codim][upface][2],[genIndex,idMat]);
            od;
            
            #replace all faces in the generator-orbit with representatives:
            if Size(orbit)>1
               then
                for o in orbit{[2..Size(orbit)]}
                  do
                    upindices:=Filtered([1..Size(hasse[codim])],
                                       i->IsSubset(hasse[codim][i][1],faces[o]));
                    
                    entry:=[genIndex,RepresentativeActionOnRightOnSets(
                                      Set(vertices{gen}),
                                      Set(vertices{faces[o]}),
                                      pgrpreps)
                            ];
                    for upface in upindices
                      do
                        AddSet(hasse[codim][upface][2],entry);
                    od;
#                    for upface in upindices
#                      do
#                        AddSet(hasse[codim][upface][3],o);
#                    od;
#
#                    faces[o]:=[genIndex,RepresentativeActionOnRightOnSets(
#                                      Set(vertices{gen}),
#                                      Set(vertices{faces[o]}),
#                                      pgrpreps)
#                               ];
                od;
            fi;
        od;
        return newhasseentry;
    end;
    
    
    #this replaces the matrices in the partial Hasse diagram with integers and
    # returns a list of matrices. This is done to comply with the structure
    # of a resolution in HAP.
    
    reformatHD:=function(hasse)
        local   elts,  faces,  face,  line;

        elts:=[];
        for faces in hasse 
          do
            for face in faces
              do
                if IsList(face[1]) and ForAll(face[1],IsInt)
                   then
                    for line in face[2]
                      do
                      Add(elts,line[2]);
                    od;
                else
                    Add(elts,face[2]);
                fi;
            od;
        od;
        if elts<>[]
           then
            Add(elts,IdentityMat(Size(elts[1])));
            elts:=Set(elts);
            for faces in hasse 
              do
                for face in faces
                  do
                    if IsList(face[1]) and ForAll(face[1],IsInt)
                       then
                        for line in face[2]
                          do
                            line[2]:=PositionSet(elts,line[2]);
                        od;
                    else
                        face[2]:=PositionSet(elts,line[2]);
                    fi;
                od;
            od;
        fi;
        return elts;
    end;
    
   
######################################################################
######################################################################

    
    
    
    ##################################################
    ## THIS IS JUST FOR BIEBERBACH GROUPS!
    ##
    ## here,  hasse has to be a list contining the generators of 
    ## dimension $k$ in the $k+1$st entry.
    ## each of the genertors already knows it's unoriented boundary.
    ##
    
    calculateBoundary:=function(k,j,hasse)
        local   dirLess,  verticesOfFace,  boundaryFromPair,  face,  
                linesdone,  linestodo,  linesinpoint,  line,  linemat,  
                points,  point,  pos,  idMat,  firstlinepos,  
                firstline,  linesforthisrun,  nextlines,  pointsdone,  
                linebound,  otherline,  boundary,  otherlinebound,  
                orientedOtherLine;
        
        dirLess:=function(face)
            return [AbsInt(face[1]),face[2]];
        end;
        
        verticesOfFace:=function(dim,elt,vertexvectors,hasse)
            local   vectors;
            vectors:=vertexvectors{hasse[dim+1][AbsInt(elt[1])]}*elt[2];
            return Set(vectors,v->Position(vertexvectors,v));
        end;


        boundaryFromPair:=function(dim,pair,hasse)
            local   gen,  bound;
            gen:=hasse[dim+1][AbsInt(pair[1])];
            bound:=gen[2];
            return List(bound,b->[SignInt(pair[1])*b[1],b[2]*pair[2]]);
        end;


        face:=hasse[k+1][j];
        if k=1    # First, define the boundary of 1-faces
           then
            if face[2][1]>face[2][2]
               then
                linesdone:=[face[2][1],[-face[2][2][1],face[2][2][2]]];
            else
                linesdone:=[[-face[2][1][1],face[2][1][2]],face[2][2]];
            fi;
        else
            #
            #now the other faces. Note that in an n-face every 
            # n-2 face is contained in exactly 2 n-1 faces.
            ##############################

            linestodo:=Set(face[2],i->dirLess(i));
            #    # the edges which are generators for this module:
            #    GenLinesTodo:=Filtered(linestodo,i->i[2]=idMat);
            #    # edges which are images under generators for this module:
            #    ImLinesTodo:=Filtered(linestodo,i->i[2]<>idMat);  


            # As the group is Bieberbach, each point has a unique name.
            # generate a list to look up the incident lines for each point:
            linesinpoint:=[];
            for line in linestodo  
              # form: line=[<pos>,<mat>]
              # <pos>: generator position
              # <mat>: matrix taking generator to actual line.
              do
                linemat:=line[2];
                points:=List(hasse[k][line[1]][2],p->[AbsInt(p[1]),p[2]*linemat]);
                for point in points 
                  #form of point as form of line but one dimension lower
                  do
                    pos:=PositionProperty(linesinpoint,pl->pl[1]=point);
                    if pos=fail
                       then
                        AddSet(linesinpoint,[point,[line]]);
                    else
                        AddSet(linesinpoint[pos][2],line);
                    fi;
                od;
            od;
            idMat:=IdentityMat(Size(hasse));
            firstlinepos:=PositionProperty(linestodo,i->i[2]=idMat);
            if firstlinepos=fail
               then
                firstline:=Remove(linestodo);
            else
                firstline:=linestodo[firstlinepos];
                Remove(linestodo,firstlinepos);
            fi;

            linesforthisrun:=[firstline];
            nextlines:=[];
            pointsdone:=[];
            linesdone:=[];

            repeat
                for line in linesforthisrun
                  do
      if k=2
         then
           Print("\n",line[1],line[2][Size(line[2])]," ->");
      fi;
                    if not ForAny([[line[1],line[2]],[-line[1],line[2]]],i-> i in linesdone)
                       then
                        AddSet(linesdone,line);
                    fi;
                    linebound:=boundaryFromPair(k-1,line,hasse);
                    points:=Difference(List(linebound,p->dirLess(p)),pointsdone);
                    for point in points
                      do
                        AddSet(pointsdone,point);
                        otherline:=First(linesinpoint,l->l[1]=point);
                        otherline:=First(otherline[2],l->l<>dirLess(line));
                        
                        if otherline in linestodo
                           then
                            RemoveSet(linestodo,otherline);
#                            Add(nextlines,otherline);
                            boundary:=boundaryFromPair(k-1,line,hasse);
                            otherlinebound:=boundaryFromPair(k-1,otherline,hasse);
                            if SignInt(First(boundary,
                                       i->[AbsInt(i[1]),i[2]]=point)[1])
                               =SignInt(First(otherlinebound,
                                       i->[AbsInt(i[1]),i[2]]=point)[1])
                               then 
                                orientedOtherLine:=[-otherline[1],otherline[2]];
                            else
                                orientedOtherLine:=otherline;
                            fi;
                            if k=2
                               then
                                Print([orientedOtherLine[1],
                                       orientedOtherLine[2][Size(orientedOtherLine[2])]]," ");
                            fi;
                            Add(linesdone,orientedOtherLine);
                            Add(nextlines,orientedOtherLine);
                        fi;
                    od;
                od;
                linesforthisrun:=StructuralCopy(nextlines);
                nextlines:=[];
            until linesforthisrun=[];
        fi;
        ## return new boundary entry:
        return Set(linesdone);
    end;

######################################################################
######################################################################


Polymake(poly,"FACE_LATTICE VERTICES");
dim:=Size(Preps[1])-1;
vertices:=Polymake(poly,"VERTICES");
MakeImmutable(vertices);
##
# Add initial (top) node:
# Note that the face lattice is generated top- down. So the 
# order has to be reversed to be ascending in dimension.
##
hasse:=Concatenation([[[[1..Size(vertices)],[]]]],
               StructuralCopy(Polymake(poly,"FACE_LATTICE"){[1..dim]}));
#    moduleGenerators:=List([1..dim+1],i->[]);

#moduleGenerators is a list of list. The i^{th} entry contains the 
# positions of the module generators of dimension i in the Hasse diagram.
#    moduleGenerators[1]:=[1];# so this means [hasse[1][1]];

#Now the next nodes:
for codim in [1..dim]
  do
    timetmp:=Runtime();
    Print(codim,": ",Size(hasse[codim+1])," \c");
    #At this stage, hasse[codim+1] is a list of faces.
    removeSomeFaces(List(hasse[codim],i->i[1]),hasse[codim+1]);
    hasse[codim+1]:=Set(hasse[codim+1]);
    Print("-> ",Size(hasse[codim+1])," \c");
    faceOrbits:=orbitDecompositionAsIndices(hasse[codim+1],vertices,Preps);
    Print(".\c");
    #        Error("break");
    hasse[codim+1]:=changeHasseEntries(faceOrbits,codim,vertices,hasse,Preps);
    Print(StringTime(Runtime()-timetmp),"\n ");
    timetmp:=Runtime();
od;
hasse:=Reversed(hasse);
for k in [1..dim]
  do
    for j in [1..Size(hasse[k+1])]
      do
        hasse[k+1][j][2]:=calculateBoundary(k,j,hasse);
    od;
od;
Print("boundary: ",StringTime(Runtime()-timetmp),"\n");
timetmp:=Runtime();

#    for dim in [1..Size(moduleGenerators)]
#      do
#        Apply(moduleGenerators[dim],i->hasse[dim][i][1]);
#    od;        
#    elts:=reformatHD(hasse);
    #    return rec(hasse:=hasse,modgens:=moduleGenerators, elts:=elts);
    elts:=reformatHD(hasse);
    Print("formating: ",StringTime(Runtime()-timetmp),"\n");
    return rec(hasse:=hasse,elts:=elts);
end);