#############################################################################
##
#W geometricjunk.gi 			 HAPcryst package		 Marc Roeder
##
##  

##
#H @(#)$Id: geometricjunk.gi, v 0.1.1 2007/08/23 18:08:51 gap Exp $
##
#Y	 Copyright (C) 2006 Marc Roeder 
#Y 
#Y This program is free software; you can redistribute it and/or 
#Y modify it under the terms of the GNU General Public License 
#Y as published by the Free Software Foundation; either version 2 
#Y of the License, or (at your option) any later version. 
#Y 
#Y This program is distributed in the hope that it will be useful, 
#Y but WITHOUT ANY WARRANTY; without even the implied warranty of 
#Y MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#Y GNU General Public License for more details. 
#Y 
#Y You should have received a copy of the GNU General Public License 
#Y along with this program; if not, write to the Free Software 
#Y Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
##
Revision.("/home/roeder/gap/HAPcryst/HAPcryst/lib/geometricjunk_gi"):=
	"@(#)$Id: geometricjunk.gi, v 0.1.1 2007/08/23   18:08:51  gap Exp $";

InstallMethod(FundamentalDomainFromRegularPointOnRight,"for crystallographic groups",
        [IsVector,IsRecord],
        function(center,groupdata)
    local   trans,  gram,  orbstab,  stabsize;
    if not (IsBound(groupdata.pointGrpReps) and IsBound(groupdata.transformation))
       then
        Error("the group data record is not valid");
    fi;
    trans:=groupdata.transformation;
    #gram:=Inverse(TransposedMat(trans)*trans);
    
    ######################################################
    #### DO SOMETHING!!!
    ######################################################
    
    gram:=GramianMatFromGroup(Group(List(groupdata.pointGrpReps,LinearPartOfAffineMatOnRight)));

    ######################################################
    ####  / DO SOMETHING!!! /
    ######################################################

    if not gram=IdentityMat(Size(groupdata.transformation))
       then
        Info(InfoHAPcryst,1,
             "Warning: gramian matrix is not the identity. Correctness of calculations is not guaranteed.");
    fi;
    orbstab:=OrbitStabilizerInUnitCubeOnRight(center,groupdata.pointGrpReps);
    stabsize:=Size(Group(orbstab.stabilizergens));
    if stabsize>1
       then
        Info(InfoHAPcryst,1,
             "this is not a general point. You may not get a fundamental domain.\nSize of stabilizer: ",stabsize);
    fi;
    return FundamentalDomainFromOrbit(gram,center,orbstab.orbit);
end);
    
    
##############################
#
#

InstallMethod(FundamentalDomainFromOrbit,"for crystallographic groups",
        [IsMatrix,IsVector,IsDenseList],
        function(gram,center,orbit)
    local   pointlist;
    pointlist:=PointlistForFundamentalDomainFromOrbit(gram,center,orbit);
    return DirichletDomain(gram,center,pointlist);
end);

##############################
#
# 
#

InstallMethod(PointlistForFundamentalDomainFromOrbit,"for crystallographic groups",
        [IsMatrix,IsVector,IsDenseList],
        function(gram,center,orbit)
    local   dim,  T,  transbasis,  boundingbox,  boundingvertices,  
            boundingverticesgram,  dirichletpoints,  centernorm,  
            trans,  orbittranslate,  centertrans,  p,  pnorm,  
            difference;

    if Size(Set(orbit,Size))<>1
       then
        Error("points must have the same dimension");
    elif Size(center)<>Size(orbit[1])
      then
        Error("dimension of center point does not match dimension of orbit points");
    fi;

    ##########
    # define a set $T$ of vectors representing the translations 
    # inside the unit cube.
    #
    dim:=Size(center);
    T:=Tuples([-1,0,1],dim);

    ##########
    # generate a bounding box given by the translations along the 
    # canonical basis. This only works, if the scalar product is
    # the standard one. More code for the other cases is to be written.
    #
    transbasis:=Set(IdentityMat(dim));
    UniteSet(transbasis,-transbasis);
    boundingbox:=DirichletDomain(gram,center,
                         Union(List(orbit,o->o+transbasis)),
                         "boundingbox");
    boundingvertices:=Polymake(boundingbox,"VERTICES");
    boundingverticesgram:=boundingvertices*gram;

    ##########
    # now calculate the translates of the orbit by elements of T
    # and keep a part of it to use with the DirichletDomain method.
    # To determine which points to keep, we do a Dirichlet-Voronoi like
    # calculation for each translate. If a translate of an orbit element
    # defines a halfspace containing a bounding vertex, it is kept.
    #
    # Otherwise it does not contribute to the construction of the Dirichlet 
    # domain (it is "too far away" from the center).
    #
    dirichletpoints:=[];
    centernorm:=center*gram*center;
    for trans in T
      do
        orbittranslate:=Set(orbit+trans);
        centertrans:=center+trans;
        for p in orbittranslate
          do
            pnorm:=p*gram*p;
            difference:=p-center;
            if ForAny(boundingverticesgram,y->centernorm-pnorm+2*(y*difference)>=0)
               then
                Add(dirichletpoints,p);
            fi;
        od;
    od;
    return Set(dirichletpoints);
end);

##############################
#
# The same as above, but without the bounding box calculation.
# If the point group is orthogonal, gramian matrix is the identity and 
# the orbit is trivial, this returns far too many points (3^n as opposed 
# to the 2^n needed)

InstallMethod(PointlistForFundamentalDomainFromOrbitNoTricks,
        [IsVector,IsDenseList],
        
        function(center,orbit)
    local   dim,  T;
    if Size(Set(orbit,Size))<>1
       then
        Error("points must have the same dimension");
    elif Size(center)<>Size(orbit[1])
      then
        Error("dimension of center point does not match dimension of orbit points");
    fi;
    
    dim:=Size(center);
    T:=Tuples([-1,0,1],dim);
    return Union(Set(orbit,i->i+T));
end);
    
    
    
##############################
#
# The Dirichlet domain of a point <center> with respect to a list <list> 
# of other points is the set of all points $x$ which are nearer to 
# <center> than to any (other) point in <list>.
#
# This method generates a file for polymake. It contains the inequalities
# defining the Dirichlet domain of <center> with respect to <list>.
#


InstallMethod(DirichletDomain,
        [IsMatrix,IsVector,IsDenseList],
        function(gram,center,list)
    return DirichletDomain(gram,center,list,"dirichlet");
end);


InstallMethod(DirichletDomain,
        [IsMatrix,IsVector,IsDenseList,IsDirectory],
        function(gram,center,list,dir)
    return DirichletDomain(gram,center,list,"dirichlet",dir);
end);


InstallMethod(DirichletDomain,
        [IsMatrix,IsVector,IsDenseList,IsString],
        function(gram,center,list,prefix)
    if not IsDirectory(HAPCRYST_TMPDIR)
       then
        InstallValue(HAPCRYST_TMPDIR,DirectoryTemporary());
    fi;
    return DirichletDomain(gram,center,list,prefix,HAPCRYST_TMPDIR);
end);


InstallMethod(DirichletDomain,
        [IsMatrix,IsVector,IsDenseList,IsString,IsDirectory],
        function(gram,center,list,prefix,dir)
    local   matrix,  centergram,  centernorm,  point,  pointgram,  
            string,  poly;
     if list=[]
       then
        Error("there must be some points");
    elif Size(Set(list,Size))<>1
      then
        Error("points must have the same dimension");
    fi;
    matrix:=[];
    centergram:=center*gram;
    centernorm:=centergram*center;
    for point in Difference(list,[center])
      do
        pointgram:=point*gram;
        Add(matrix,Concatenation([-(centernorm-pointgram*point)],2*(centergram-pointgram)));
    od;
    string:=ConvertMatrixToPolymakeString("INEQUALITIES",matrix);
    poly:=CreatePolygon(prefix,dir);
    AppendToPolygon(poly,string);
    return poly;
end);



##############################
# 
# Given an inequality which defines a hyperplane and a point <x> outside this
# hyperplane, the image of <x> under the reflection on the hyperplane is 
# returned.
#
InstallMethod(ReflectedPointGramInverse,
        [IsVector,IsVector,IsMatrix],
        function(x,inequality,graminverse)
    local   dimplusone;
    dimplusone:=Size(inequality);
    return x-1/2*inequality{[2..dimplusone]}*graminverse;
end);


InstallMethod(ReflectedPoint,
        [IsVector,IsVector],
        function(x,inequality)
    local   perpend;
    perpend:=inequality{[2..Size(inequality)]};
    return x-2/perpend^2*(perpend-inequality[1])*perpend;
end);


InstallMethod(ReflectedPoint,
        [IsVector,IsVector,IsMatrix],
        function(x,inequality,gram);
    return ReflectedPointGramInverse(x,inequality,Inverse(gram));
end);




##############################
#
# Find all points in a list which lie on the far side of a 
# point <x> (as seen from the point <center>
#

#InstallMethod(PointsInsideCenteredBox,
#        [IsVector,IsVector,IsMatrix,IsList],
#        function(center,x,gram,list)
#    
    