#############################################################################
##
#W hasse.gi 			 HAPcryst package		 Marc Roeder
##
##  

##
#H @(#)$Id: hasse.gi, v 0.1.1 2007/08/23 18:08:51 gap Exp $
##
#Y	 Copyright (C) 2006 Marc Roeder 
#Y 
#Y This program is free software; you can redistribute it and/or 
#Y modify it under the terms of the GNU General Public License 
#Y as published by the Free Software Foundation; either version 2 
#Y of the License, or (at your option) any later version. 
#Y 
#Y This program is distributed in the hope that it will be useful, 
#Y but WITHOUT ANY WARRANTY; without even the implied warranty of 
#Y MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#Y GNU General Public License for more details. 
#Y 
#Y You should have received a copy of the GNU General Public License 
#Y along with this program; if not, write to the Free Software 
#Y Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
##
Revision.("/home/roeder/gap/HAPcryst/HAPcryst/lib/hasse_gi"):=
	"@(#)$Id: hasse.gi, v 0.1.1 2007/08/23   18:08:51  gap Exp $";

####################
# HyperplaneForConeProjection takes a vector <point> and a list <vertices> 
# of vectors and finds a hyperplane which does not contain <point> and 
# is not parallel to any of the lines defined by <vertices>-<point>.
#
# This method is not guaranteed to find a hyperplane. If it doesn't an
# error is issued.
#
# This hyperplane can be used to do a Fourier-Motzkin like Projection to find
# the (partial) Hasse diagram of the polyhedron given by the union of 
# <vertices> and {<point>}.
#
# Note that it is not checked, wheather <point> is in <vertices> or not.
# If a symmetric positive definite matrix <HAPcrystGram> is passed as an option,
# this is used as the gramian matrix of the inner product.
#
# The returned value is a vector <v> representing the hyperplane defined by
# x*v{[2..Size(v)]}=-v[1]
# NOTE! The vector v{[2..Size(v)]} is generally not perpendicular to the 
# hyperplane.
# The hyperplane is defined by an EQUATION. This just happens to be the
# same if the gramian matrix is the identity matrix.
#

######################################################################
#
# 
InstallMethod(HyperplaneForConeProjection,[IsVector,IsDenseList],
        function(point,vertices)

    local   gram,  directions,  average,  directionProdList,  Wrongs,  
            maxwrong,  w,  lambda,  rightside,  v;

    if not ForAll(vertices,IsVector)
       then
        Error("vertex list must be a list of vectors");
    fi;
    gram:=ValueOption("HAPcystGram");
    if gram=fail
       then
        gram:=IdentityMat(Size(point));
    fi;
    directions:=List(Filtered(vertices,v->v<>point),v->v-point);
    ##
    # first find a hyperplane which has the "right direction"
    # This is not nice!
    #
    average:=Sum(directions);

    directionProdList:=List(directions,d->[d,(average*gram)*d]);
    if Number(directionProdList,i->SignRat(i[2])=-1)>Number(directionProdList,i->SignRat(i[2])=1)
       then
        average:=-average;
        Apply(directionProdList,i->[i[1],-i[2]]);
    fi;
    Wrongs:=Filtered(directionProdList,i->SignRat(i[2])<>1);
    if Wrongs<>[]
       then
        Info(InfoHAPcryst,1,"First choice of hyperplane wasn't good.\n  Cross your fingers that the iterative change algorithm works.");
    fi;
    while Wrongs<>[]
      do
        Print("*\c");
        maxwrong:=Maximum(List(Wrongs,w->-w[2]));
        w:=First(Wrongs,w->w[2]=-maxwrong)[1];
        if not maxwrong=0
           then
            lambda:=-((w*gram)*average)/((w*gram)*w)*11/10;#is adding 10% a good idea?
        else
            lambda:=1/10;
        fi;
        average:=average+lambda*w;
        directionProdList:=List(directions,d->[d,(average*gram)*d]);
        Wrongs:=Filtered(directionProdList,i->SignRat(i[2])<>1);
    od;

    #    orthogonalDirections:=Filtered(directions,i->(average*gram)*i=0);
    #    if orthogonalDirections<>[]
    #       then
    #        Info(InfoHAPcryst,1,"hyperplane is parallel to one of the directions. Changing it blindly.\n Cross your fingers!\n");
    #        scale:=Maximum(orthogonalDirections,i->i*i)*1/10;
    #        average:=Sum(Concatenation([[average],scale*orthogonalDirections]));
    #    fi;
    ##
    # if needed, move it, such that it does not contain <point>
    #
    if (average*gram)*point=0
       then
        Info(InfoHAPcryst,1,"shifting hyperplane (away from cone point)\n");
        rightside:=(point*gram)*point;
        if rightside=0
           then
            rightside:=1+Maximum(List(vertices,x->(x*gram)*average));
        fi;
    else
        rightside:=0;
    fi;
    ##
    # finally test if the (messy) choice really works:
    #
    if not Size(Set(directions,d->SignRat((d*gram)*average)))=1
       then
        Error("cannot find hyperplane. Something is wrong. Some more intelligence needed");
    fi;
    v:=[-rightside];
    Append(v,average);
    return v;
end);

##############################

InstallMethod(WhichSideOfHyperplane,
        [IsVector,IsVector],
        function(vector,inequality)
    local   eval;
    if not Size(vector)+1=Size(inequality)
       then
        Error("<vector> and <inequality> must be given by a vector.");
    fi;
    if not ForAll(Union(vector,inequality),IsRat)
       then
        Error("This does only work for rational vector spaces");
    fi;
    eval:=vector*inequality{[2..Size(inequality)]}+inequality[1];
    if eval>0
       then
        return 1;
    elif eval<-1
      then
        return -1;
    elif eval=0
      then
        return 0;
    else
        return fail;
    fi;
end);

##############################    
# the projection $p+\lambda x$ of $x$ onto the hyperplane given by 
# the equation v satisfies:
#
# \sum_{i=2}^n v_i(p_i+\lambda x_i)=-v_1
#  which used as as
# \lambda \sum_{i=2}^n v_ix_i=-v_i -\sum v_ip_i
#

InstallMethod(ProjectionToAffineSpace,
        [IsVector,IsVector,IsVector],
        function(vertex,point,equation)
    local   rightside,  direction,  leftside,  returnval;

    if not (Size(vertex)+1=Size(equation) and Size(point)=Size(vertex))
       then
        Error("<direction> and <equation> must be given by a vector.");
    fi;
    rightside:=-equation[1]-equation{[2..Size(equation)]}*point;
    direction:=vertex-point;
    leftside:=direction*equation{[2..Size(equation)]};
    if leftside<>rightside
       then
        if leftside=0
           then
            Error("can't project onto this plane");
        else
            returnval:=(rightside/leftside*direction)+point;
        fi;
    else
        returnval:=vertex;  #=direction+point 
    fi;
    if not WhichSideOfHyperplane(returnval,equation)=0
       then
        Error("Hier kann wohl jemand nicht programmieren...");
    fi;
    return returnval;
end);


##############################
# This needs POINTS to be known to the polyhedron. It will blow up otherwise.
#
# calculate a  polyhedron from a given one <poly>.
# The new polyhedron is one dimension smaller and calculated by
# a projection via ProjectToAffineSpace and a call of polymake to
# get the vertices.
# A starting point for the projection has to be chosen.
# If the option "HAPcrystGram" is set, it's value is taken for the gramian
# matrix of the inner product.

InstallMethod(ProjectionOfPolygonByCone,"For Polygons",
        [IsVector,IsRecord],
        function(point,poly)
    local   tmpPoly,  points,  planeToProjectTo,  imagepoints,  
            projection,  conepoint;
    if not IsDirectory(HAPCRYST_TMPDIR)
       then
        InstallValue(HAPCRYST_TMPDIR,DirectoryTemporary());
    fi;
    tmpPoly:=CreatePolygon("tmpProjectionByCone",HAPCRYST_TMPDIR);
    Polymake(poly,"POINTS VERTICES");
    points:=Set(poly.POINTS);
    if not point in poly.VERTICES
       then
        Error("<point> must be a vertex of <poly>. Sorry.");
    fi;
    RemoveSet(points,point);
    planeToProjectTo:=HyperplaneForConeProjection(point,points);

    imagepoints:=List(points,p->ProjectionToAffineSpace(p,point,planeToProjectTo));

    AppendPointlistToPolygon(tmpPoly,Unique(Compacted(imagepoints)));
    #    imagevertices:=Polymake(tmpPoly,"VERTICES");

    #    inverseProjection:=Set(imagevertices,
    #                           i->[i,
    #                               vertices[Position(imagepoints,i)]
    #                               ]);
    projection:=List([1..Size(imagepoints)],
                     i->[points[i],imagepoints[i]]);
    return rec(projection:=projection,
               conepoint:=point,
               poly:=tmpPoly
               );
end);


