#############################################################################
##
#W resolutionBieberbach.gi 			 HAPcryst package		 Marc Roeder
##
##  

##
#H @(#)$Id: resolutionBieberbach.gi, v 0.1.4 2007/10/17 14:22:33 gap Exp $
##
#Y	 Copyright (C) 2006 Marc Roeder 
#Y 
#Y This program is free software; you can redistribute it and/or 
#Y modify it under the terms of the GNU General Public License 
#Y as published by the Free Software Foundation; either version 2 
#Y of the License, or (at your option) any later version. 
#Y 
#Y This program is distributed in the hope that it will be useful, 
#Y but WITHOUT ANY WARRANTY; without even the implied warranty of 
#Y MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#Y GNU General Public License for more details. 
#Y 
#Y You should have received a copy of the GNU General Public License 
#Y along with this program; if not, write to the Free Software 
#Y Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
##
Revision.("/home/roeder/gap/HAPcryst/HAPcryst/lib/resolutionBieberbach_gi"):=
	"@(#)$Id: resolutionBieberbach.gi, v 0.1.4 2007/10/17   14:22:33  gap Exp $";
#############################################################################
##
#O ResolutionBieberbachGroup(<group>)
##
InstallMethod(ResolutionBieberbachGroup,
        [IsGroup],
        function(group)
    return ResolutionBieberbachGroup(group,List([1..Size(Representative(group))-1],i->0));
end);

#############################################################################
##
#O ResolutionBieberbachGroup(<group>,<center>)
##
InstallMethod(ResolutionBieberbachGroup,"for affine cryst groups on right",
        [IsGroup,IsVector],
        function(group,center)
    local   gram,  poly,  fl;
    
    gram:=GramianOfAverageScalarProductFromFiniteMatrixGroup(PointGroup(group));
    if not Order(gram)=1
       then
        Info(InfoHAPcryst,2,"non standard");
    fi;
    poly:=FundamentalDomainBieberbachGroup(center,group,gram);
    fl:=FaceLatticeAndBoundaryBieberbachGroup(poly,group);
    return ResolutionFromFLandBoundary(fl, group);
end);


#############################################################################
##
#O ResolutionFromFLandBoundary(<fl>,<group>)
##
InstallMethod(ResolutionFromFLandBoundary,"for affine cryst groups on right",
        [IsRecord,IsGroup],
        function(fl,group)
    local   dimension,  boundary,  homotopy,  properties;

    dimension:=function(k)
        if k<0 or k>Size(fl.hasse)-1
           then
            return 0;
        else
            return Size(fl.hasse[k+1]);
        fi;
    end;
    
    boundary:=function(k,j)
        if k<=0 or k>=Size(fl.hasse)
           then
            return [];
        else
            return fl.hasse[k+1][j][2];
        fi;
    end;
    
    if not (IsAffineCrystGroupOnRight(group) and IsStandardSpaceGroup(group))
       then
        Error("group is not a StandardSpaceGroup acting on right");
    fi;
    if not ForAll(fl.elts,i->i in group)
       then
        Error("group does not match face lattice");
    fi;
    
    return Objectify(HapResolution,
                   rec(
                       dimension:=dimension,
                       boundary:=boundary,
                       homotopy:=fail,
                       elts:=fl.elts,
                       group:=group,
                       hasse:=fl.hasse,
                       properties:=
                       [["length",Size(fl.hasse)-1],
                        ["type","resolution"],
                        ["characteristic",0]
                        ]
                       ));
end);



